# -*- coding: utf-8 -*-

"""XML Utilities.

XML scripts in Python.

 * ppx: pretty print XML source in human readable form.
 * xp: use XPath expression to select nodes in XML source.
 * validate: validate XML source with XSD or DTD.
 * transform: transform XML source with XSLT.
"""


# Xul version.
__version_info__ = ('2', '4', '1')
__version__ = '.'.join(__version_info__)
